
/**
 * Copyright (C) 2010-2014 Ikanos Communications.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * version 2 as published by the Free Software Foundation. 
 *
 * Info : BCH decode algorithm for Fusiv VX185 on-chip NAND 
 *        flash controller driver
 */

#include <fusiv_nand.h>
//#include <bch_decode.h>
#include <vx185_scu.h>

#define MATRIX_A(x) pMatrix_a[x]
#define MATRIX_B(x) pMatrix_b[x]
#define MATRIX_C(x) pMatrix_c[x]
#define ELEMENT(x) pElement[x]
#define SYNVAL(x) syn_val[x]
#define SET_MATRIX_A(x,y) pMatrix_a[x] = (y)
#define SET_MATRIX_B(x,y) pMatrix_b[x] = (y)
#define SET_MATRIX_C(x,y) pMatrix_c[x] = (y)
#define SET_ELEMENT(x,y) pElement[x] = (y)
#define SET_SYNVAL(x,y) syn_val[x] = (y)

#define ERRLOC_SIZEX (18)

#define A2I_TABLESIZE (8192)
static u16 a_to_i[A2I_TABLESIZE];

static const short int mask = 1 << 12;
// p(x) = x^10 + x^8 + x^7 + x^5 + x^3 +x^2 + x + 1
static const short int polyterms = 0x5af;
static const short int polydegree = 13;


static int errloc_sizex;
static short *perr_loc_ply;

short *err_loc_ply(int x, int y) {
	return perr_loc_ply + ((x * errloc_sizex) + y); 
}

/* Read BCH error flags */
static inline unsigned long bch_get_error_flags(void)
{
	return fusiv_nfc_readl(NFC_BCH_STATUS);
}

/* Read Syndrome values for error location detection */
static void read_syndrome_vals(unsigned short *psynvals, int x)
{
        /* Get the 8 hardware generated syndrome values relating to the
        *  Xth 512 bytes of the page just read.
        */
        unsigned long synregval;
        int y;

        for (y = 0; y < NR_SYNDROMES/2; y++) {
                synregval = fusiv_nfc_readl(NFC_SECTx_SYNDROMEy(x, y));
                *psynvals++ = (unsigned short)synregval;
                *psynvals++ = (unsigned short)((synregval >> 16) & 0xffff);
        }
}

void print_syndrome_registers (void) {

	uint8_t i;

	printk("%s BCH_Status: 0x%x, INT_Status: 0x%x, \n",__FUNCTION__, \
		fusiv_nfc_readl(NFC_BCH_STATUS),fusiv_nfc_readl(NFC_INT_STATUS));

	for (i = 0; i < 16; i++) {
		printk("Section-%d, Syndrome-reg-%d: 0x%x\n", \
			i/4, i%4, fusiv_nfc_readl(NFC_SECT0_SYNDROME12 + i * 4));
	}
}

/******************************************************************************
 * BCH syndrome support routines for NAND booting.
 * -----------------------------------------------
 *
 * BCH detection and correction algorithms normally use precalculated forward
 * and reverse lookup tables to speed up their calculations.  These tables are
 * too big to fit in the bootrom (typically about 16kB each) so we cannot use
 * them as they are and calculating the entries on-demand takes too long.
 *
 * We therefore comprimise and have a 4kB table for the forward lookup
 * and an 8kB for the reverse lookup.  The forward lookup table provides
 * every fourth entry of the full table and the code generates the
 * intermediate entries by iterating from the table value up to three more
 * times.
 *
 * The reverse table ("what offset in the forward table holds value x")
 * provides the most significant 8 bits of the 13 bit offset into the
 * forward table (i.e. the start of a block of 32 entries in the forward table
 * where the value is).  The code then searches these 32 entries in the forward
 * table until it locates the value and hence the full offset required.
 *
 *******************************************************************************/

static void gen_a2i(void)
{
    /* Generate all even entries in the a2i table and store the table at the
     *  start of LMEM.
     */
    int i;
    unsigned short *p;
    short alpha;

    p = a_to_i;
    alpha = 1;
    for (i = 0; i < A2I_TABLESIZE; i++) {
        if ((i & 1) == 0) {
            *p++ = alpha;
        }
        if (alpha >= mask) {
            alpha = polyterms ^ ((alpha ^ mask) << 1);
        } else {
            alpha <<= 1;
        }
    }
}

static short a2i(short i)
{
    short alpha, table_index;

    table_index = i >> 1;
    alpha = a_to_i[table_index];
    if ((i & 1) != 0) {
        /* Table only contains even entries, we must calculate odd values */
        if (alpha >= mask) {
            alpha = polyterms ^ ((alpha ^ mask) << 1);
        } else {
            alpha <<= 1;
        }
        if (i == 8191) {
            return 0;
        }
    }
    return alpha;
}

static const u8 i_to_a[8192] = {
    0, 0, 0, 248, 0, 241, 248, 110, 0, 103, 241, 170, 248, 47, 111, 233,
    0, 226, 103, 45, 241, 221, 170, 40, 248, 235, 47, 96, 111, 162, 233, 87,
    0, 79, 226, 158, 103, 143, 45, 155, 241, 14, 221, 228, 170, 88, 40, 25,
    248, 25, 235, 218, 47, 37, 96, 134, 111, 32, 162, 94, 233, 189, 87, 214,
    0, 207, 79, 239, 226, 84, 158, 182, 103, 90, 143, 25, 45, 87, 155, 90,
    241, 95, 14, 17, 222, 211, 228, 61, 170, 126, 88, 26, 40, 197, 25, 30,
    248, 113, 25, 72, 235, 151, 218, 227, 47, 147, 37, 76, 96, 250, 134, 136,
    111, 18, 32, 246, 162, 82, 94, 81, 233, 156, 189, 7, 87, 220, 214, 217,
    0, 210, 207, 130, 79, 245, 239, 213, 226, 157, 84, 148, 158, 255, 182, 183,
    103, 57, 90, 10, 143, 238, 25, 136, 45, 73, 87, 134, 155, 164, 90, 74,
    241, 205, 95, 105, 14, 64, 17, 240, 222, 220, 211, 43, 228, 215, 61, 143,
    170, 128, 126, 44, 88, 27, 26, 242, 40, 232, 198, 140, 25, 69, 30, 8,
    248, 165, 113, 199, 25, 231, 72, 92, 235, 174, 151, 146, 218, 136, 227, 76,
    47, 83, 147, 6, 37, 228, 76, 79, 96, 1, 250, 82, 134, 18, 136, 125,
    111, 22, 18, 149, 32, 98, 246, 190, 162, 13, 82, 119, 94, 18, 81, 238,
    233, 134, 156, 88, 189, 10, 7, 219, 87, 53, 220, 254, 214, 85, 217, 203,
    0, 196, 210, 105, 207, 90, 130, 78, 79, 195, 245, 46, 239, 247, 213, 142,
    226, 187, 157, 127, 84, 80, 148, 66, 158, 211, 255, 193, 182, 162, 183, 2,
    103, 237, 57, 15, 90, 142, 10, 224, 143, 183, 238, 76, 25, 233, 136, 90,
    45, 231, 73, 184, 87, 82, 134, 11, 155, 115, 164, 6, 90, 111, 74, 54,
    241, 215, 205, 157, 95, 192, 105, 57, 14, 85, 64, 193, 17, 53, 240, 224,
    222, 69, 220, 181, 211, 174, 43, 129, 228, 101, 215, 167, 61, 138, 143, 55,
    170, 118, 128, 4, 126, 52, 44, 10, 88, 114, 27, 249, 26, 75, 242, 11,
    40, 125, 232, 75, 198, 255, 140, 234, 25, 72, 69, 81, 30, 72, 8, 221,
    248, 137, 165, 202, 113, 123, 199, 223, 25, 206, 231, 44, 72, 131, 92, 237,
    235, 175, 174, 52, 151, 8, 146, 248, 218, 191, 136, 150, 227, 141, 76, 249,
    47, 67, 83, 99, 147, 172, 6, 157, 37, 154, 228, 66, 77, 127, 79, 103,
    96, 41, 1, 49, 250, 3, 83, 206, 134, 128, 18, 24, 136, 41, 125, 231,
    111, 0, 22, 208, 18, 62, 149, 62, 33, 94, 98, 225, 246, 132, 190, 15,
    162, 132, 13, 121, 82, 37, 119, 195, 94, 235, 18, 195, 81, 171, 238, 20,
    233, 48, 134, 198, 156, 98, 88, 58, 189, 232, 10, 73, 7, 201, 219, 57,
    87, 136, 53, 132, 220, 17, 254, 207, 214, 99, 85, 212, 217, 35, 203, 201,
    0, 193, 196, 87, 210, 169, 105, 28, 207, 129, 90, 92, 130, 205, 78, 112,
    79, 60, 195, 129, 245, 125, 46, 30, 239, 200, 247, 166, 213, 236, 142, 9,
    226, 36, 187, 40, 157, 190, 127, 37, 84, 50, 80, 83, 148, 34, 66, 91,
    158, 49, 212, 106, 255, 196, 193, 194, 182, 117, 162, 225, 183, 66, 2, 188,
    103, 231, 237, 249, 57, 201, 15, 20, 90, 54, 142, 2, 10, 221, 224, 54,
    143, 8, 183, 74, 238, 203, 76, 125, 25, 214, 233, 86, 136, 217, 90, 30,
    45, 12, 231, 247, 73, 166, 184, 164, 87, 73, 82, 227, 134, 187, 11, 74,
    155, 1, 115, 124, 164, 113, 6, 251, 90, 188, 111, 9, 74, 29, 54, 29,
    241, 70, 215, 129, 205, 195, 157, 103, 95, 215, 192, 20, 105, 93, 57, 115,
    14, 230, 85, 4, 64, 43, 193, 124, 17, 46, 53, 198, 240, 36, 224, 124,
    222, 241, 69, 41, 220, 165, 182, 133, 211, 4, 174, 183, 43, 142, 129, 155,
    228, 41, 101, 168, 215, 45, 167, 161, 61, 240, 138, 209, 143, 222, 55, 1,
    170, 223, 118, 19, 128, 74, 4, 34, 126, 181, 52, 121, 44, 17, 10, 6,
    88, 16, 114, 34, 27, 42, 249, 245, 26, 199, 75, 71, 242, 153, 11, 251,
    40, 109, 125, 59, 232, 91, 75, 254, 198, 149, 255, 59, 140, 161, 234, 164,
    25, 96, 72, 196, 69, 91, 81, 119, 30, 209, 72, 147, 8, 58, 221, 114,
    248, 194, 137, 189, 165, 97, 202, 101, 113, 70, 123, 138, 199, 160, 223, 82,
    25, 135, 206, 170, 231, 137, 44, 239, 72, 155, 131, 188, 92, 38, 237, 227,
    235, 251, 175, 237, 174, 138, 53, 154, 151, 179, 8, 204, 146, 186, 248, 87,
    218, 206, 191, 180, 136, 119, 150, 252, 227, 59, 141, 93, 76, 119, 249, 73,
    47, 47, 67, 70, 83, 112, 99, 104, 147, 187, 172, 108, 6, 254, 157, 11,
    37, 154, 154, 224, 228, 177, 66, 31, 77, 3, 127, 184, 79, 123, 103, 75,
    96, 16, 41, 230, 1, 7, 49, 60, 250, 216, 3, 141, 83, 154, 206, 134,
    134, 83, 128, 56, 18, 95, 24, 226, 136, 131, 41, 175, 125, 69, 231, 187,
    111, 213, 0, 230, 22, 203, 208, 65, 18, 241, 62, 65, 149, 74, 62, 231,
    33, 141, 94, 118, 98, 68, 225, 234, 246, 226, 132, 100, 190, 143, 15, 247,
    162, 168, 132, 110, 13, 252, 121, 5, 82, 3, 37, 27, 119, 38, 196, 45,
    94, 3, 235, 108, 18, 38, 195, 68, 81, 175, 171, 107, 238, 242, 20, 12,
    233, 245, 48, 208, 134, 150, 198, 99, 156, 50, 98, 35, 88, 207, 58, 184,
    189, 217, 233, 19, 10, 244, 73, 46, 7, 144, 201, 77, 219, 185, 57, 164,
    87, 47, 136, 255, 53, 246, 132, 131, 221, 204, 17, 93, 254, 160, 207, 73,
    214, 250, 99, 61, 85, 174, 212, 168, 217, 121, 35, 2, 203, 139, 201, 166,
    0, 159, 193, 146, 196, 61, 87, 132, 210, 115, 169, 114, 105, 250, 28, 247,
    207, 31, 129, 242, 90, 54, 92, 152, 130, 160, 205, 178, 78, 111, 112, 167,
    79, 135, 60, 40, 195, 248, 129, 66, 245, 123, 125, 30, 46, 234, 30, 239,
    239, 66, 200, 152, 247, 132, 167, 152, 213, 178, 236, 197, 142, 86, 9, 200,
    226, 29, 36, 238, 187, 200, 40, 86, 157, 92, 190, 135, 127, 214, 38, 142,
    84, 177, 50, 153, 80, 51, 83, 200, 148, 145, 34, 42, 66, 28, 91, 9,
    158, 156, 49, 180, 212, 22, 106, 178, 255, 210, 196, 136, 193, 70, 194, 213,
    182, 228, 117, 209, 162, 12, 225, 197, 183, 39, 66, 27, 2, 58, 188, 236,
    103, 156, 231, 206, 237, 223, 249, 248, 57, 57, 201, 207, 15, 61, 20, 195,
    90, 224, 54, 11, 142, 78, 2, 66, 10, 10, 221, 234, 224, 57, 54, 129,
    143, 240, 8, 242, 183, 213, 74, 135, 238, 190, 203, 219, 76, 92, 125, 79,
    25, 155, 214, 134, 233, 110, 86, 40, 136, 227, 217, 81, 90, 182, 30, 61,
    45, 5, 12, 68, 231, 65, 247, 234, 73, 131, 166, 167, 184, 99, 164, 46,
    87, 154, 73, 252, 82, 101, 227, 239, 134, 60, 187, 225, 11, 104, 74, 31,
    155, 133, 1, 161, 115, 103, 124, 124, 164, 254, 113, 119, 6, 34, 251, 245,
    90, 37, 188, 194, 111, 28, 9, 30, 74, 163, 29, 251, 54, 19, 29, 125,
    241, 50, 70, 187, 215, 181, 129, 54, 205, 94, 195, 85, 157, 174, 103, 90,
    95, 75, 215, 26, 192, 12, 20, 152, 105, 96, 93, 63, 57, 131, 115, 92,
    14, 220, 230, 62, 85, 50, 4, 31, 64, 20, 43, 148, 193, 180, 124, 207,
    17, 104, 46, 127, 54, 162, 198, 243, 240, 232, 36, 227, 224, 78, 124, 130,
    222, 65, 241, 118, 69, 189, 41, 111, 220, 63, 165, 51, 182, 86, 133, 78,
    211, 123, 4, 199, 174, 173, 184, 167, 43, 244, 143, 173, 129, 71, 155, 112,
    228, 37, 41, 243, 101, 229, 168, 160, 215, 147, 45, 150, 167, 126, 161, 130,
    61, 80, 240, 169, 138, 48, 209, 178, 143, 93, 222, 171, 55, 197, 1, 205,
    170, 180, 223, 56, 118, 108, 19, 61, 128, 106, 74, 124, 4, 168, 34, 196,
    126, 184, 181, 75, 52, 49, 121, 132, 44, 218, 17, 246, 10, 44, 6, 87,
    88, 5, 16, 9, 114, 222, 34, 159, 27, 53, 42, 183, 249, 158, 245, 0,
    26, 127, 199, 169, 75, 169, 71, 146, 242, 152, 153, 209, 11, 133, 251, 194,
    40, 158, 109, 39, 125, 62, 59, 250, 232, 97, 91, 128, 75, 212, 254, 105,
    198, 4, 149, 71, 255, 229, 59, 247, 140, 243, 161, 179, 234, 100, 164, 28,
    25, 67, 96, 140, 72, 68, 196, 116, 69, 97, 91, 252, 81, 176, 119, 210,
    30, 77, 209, 146, 72, 216, 147, 114, 8, 23, 58, 152, 221, 56, 114, 169,
    248, 212, 194, 186, 137, 80, 189, 12, 165, 21, 97, 21, 203, 121, 101, 162,
    113, 104, 70, 27, 123, 217, 138, 197, 199, 120, 160, 122, 223, 84, 82, 225,
    25, 2, 135, 109, 206, 15, 170, 229, 231, 115, 137, 192, 44, 159, 239, 182,
    72, 99, 155, 53, 131, 121, 188, 209, 92, 23, 38, 163, 237, 63, 227, 117,
    235, 181, 251, 211, 175, 89, 237, 58, 175, 152, 138, 110, 53, 218, 154, 2,
    151, 112, 179, 42, 8, 98, 204, 236, 146, 186, 186, 31, 248, 104, 87, 188,
    218, 192, 206, 28, 191, 33, 180, 39, 136, 30, 119, 219, 150, 148, 252, 183,
    227, 83, 59, 183, 141, 204, 93, 27, 76, 228, 119, 43, 249, 76, 73, 66,
    48, 22, 47, 109, 67, 133, 70, 22, 83, 192, 112, 181, 99, 2, 104, 212,
    147, 70, 187, 249, 172, 117, 108, 178, 6, 244, 254, 48, 157, 166, 11, 106,
    37, 133, 154, 5, 154, 240, 224, 156, 228, 156, 177, 29, 66, 135, 31, 159,
    77, 67, 3, 158, 127, 5, 184, 180, 80, 36, 123, 65, 103, 220, 75, 49,
    96, 109, 16, 223, 41, 241, 230, 70, 1, 12, 7, 231, 49, 36, 60, 193,
    250, 47, 216, 245, 3, 168, 141, 213, 83, 16, 154, 47, 206, 251, 135, 194,
    134, 22, 83, 164, 128, 205, 56, 210, 18, 113, 95, 206, 24, 79, 226, 255,
    136, 48, 131, 0, 41, 67, 175, 137, 125, 117, 69, 215, 231, 237, 187, 196,
    111, 106, 213, 32, 0, 60, 230, 51, 22, 216, 203, 202, 208, 140, 65, 80,
    18, 11, 241, 88, 62, 189, 65, 200, 149, 112, 74, 201, 62, 191, 231, 83,
    33, 166, 141, 102, 94, 52, 118, 177, 98, 246, 68, 33, 225, 253, 234, 84,
    246, 157, 226, 6, 132, 14, 100, 153, 190, 137, 143, 142, 15, 52, 247, 50,
    162, 48, 168, 216, 132, 11, 110, 28, 13, 26, 252, 79, 121, 204, 5, 66,
    82, 255, 3, 17, 37, 81, 27, 9, 119, 95, 38, 174, 196, 113, 45, 91,
    94, 244, 3, 100, 235, 177, 108, 145, 18, 77, 38, 191, 195, 64, 68, 148,
    81, 6, 175, 9, 171, 26, 107, 42, 239, 238, 242, 107, 20, 116, 12, 34,
    233, 178, 245, 63, 48, 122, 208, 200, 134, 95, 150, 193, 198, 166, 99, 188,
    156, 108, 50, 114, 98, 24, 35, 86, 88, 39, 207, 208, 58, 13, 184, 40,
    189, 117, 217, 171, 233, 35, 19, 29, 10, 218, 244, 39, 73, 191, 46, 226,
    7, 172, 144, 222, 201, 253, 77, 238, 219, 116, 185, 55, 57, 165, 164, 36,
    87, 249, 47, 88, 136, 18, 255, 214, 53, 75, 247, 233, 132, 201, 131, 127,
    221, 187, 205, 34, 17, 160, 93, 142, 254, 154, 160, 186, 207, 159, 74, 38,
    214, 70, 250, 234, 99, 33, 61, 92, 85, 126, 174, 180, 212, 176, 168, 158,
    217, 147, 121, 159, 35, 79, 2, 135, 204, 64, 139, 253, 201, 176, 166, 191,
    0, 183, 159, 9, 194, 209, 146, 169, 196, 124, 61, 56, 87, 245, 132, 75,
    210, 252, 115, 140, 169, 151, 114, 146, 105, 128, 250, 39, 28, 179, 247, 71,
    207, 148, 31, 62, 130, 227, 242, 127, 90, 85, 54, 187, 92, 63, 152, 26,
    130, 150, 160, 243, 205, 171, 178, 169, 78, 51, 111, 118, 112, 173, 167, 199,
    79, 219, 135, 242, 60, 81, 40, 134, 195, 207, 248, 206, 129, 234, 66, 11,
    245, 119, 124, 161, 125, 251, 30, 194, 46, 167, 234, 67, 30, 225, 239, 252,
    239, 30, 66, 40, 200, 197, 152, 152, 247, 114, 132, 146, 167, 178, 152, 242,
    213, 136, 178, 180, 236, 27, 197, 209, 142, 135, 86, 238, 9, 42, 200, 153,
    226, 39, 29, 171, 36, 55, 238, 222, 188, 193, 200, 63, 40, 208, 86, 114,
    157, 180, 92, 234, 190, 253, 135, 159, 127, 233, 214, 88, 38, 186, 142, 34,
    84, 33, 177, 102, 50, 142, 153, 5, 80, 202, 51, 32, 83, 200, 200, 88,
    148, 191, 145, 100, 34, 107, 42, 9, 66, 79, 28, 216, 91, 174, 9, 16,
    158, 28, 156, 5, 49, 65, 180, 158, 212, 181, 22, 109, 106, 48, 178, 249,
    255, 206, 210, 164, 196, 215, 136, 0, 193, 230, 70, 223, 194, 47, 213, 245,
    182, 192, 229, 109, 117, 163, 209, 53, 162, 21, 12, 186, 225, 122, 197, 27,
    183, 219, 39, 28, 66, 43, 27, 183, 2, 110, 58, 210, 188, 31, 236, 42,
    103, 66, 156, 99, 231, 24, 206, 49, 237, 44, 223, 202, 249, 150, 248, 52,
    57, 73, 57, 198, 201, 212, 207, 132, 15, 225, 62, 208, 20, 195, 195, 121,
    90, 76, 224, 15, 54, 6, 11, 184, 142, 46, 78, 105, 2, 193, 66, 126,
    10, 249, 10, 4, 221, 81, 234, 75, 224, 193, 57, 157, 54, 167, 129, 181,
    143, 43, 240, 105, 8, 140, 242, 44, 183, 148, 213, 130, 74, 134, 135, 10,
    238, 119, 190, 149, 203, 254, 219, 88, 76, 146, 92, 199, 125, 82, 79, 6,
    25, 228, 155, 158, 214, 94, 134, 218, 233, 169, 110, 248, 86, 96, 40, 44,
    136, 76, 227, 72, 217, 7, 81, 246, 90, 25, 182, 239, 30, 26, 61, 17,
    45, 27, 5, 110, 12, 107, 68, 108, 231, 64, 65, 230, 247, 100, 234, 118,
    73, 93, 131, 255, 166, 1, 167, 61, 184, 35, 99, 208, 164, 77, 46, 19,
    87, 204, 154, 237, 73, 93, 252, 180, 82, 138, 101, 188, 227, 188, 239, 170,
    134, 141, 60, 230, 187, 175, 226, 56, 11, 108, 104, 70, 74, 184, 31, 223,
    155, 183, 133, 41, 1, 208, 161, 168, 115, 20, 103, 129, 124, 198, 124, 4,
    164, 59, 254, 59, 113, 147, 119, 196, 6, 121, 34, 19, 251, 71, 245, 34,
    90, 83, 37, 40, 188, 225, 194, 106, 111, 92, 28, 87, 9, 166, 30, 128,
    74, 227, 163, 247, 29, 9, 251, 124, 54, 2, 19, 249, 29, 86, 125, 74,
    241, 3, 50, 204, 70, 179, 187, 216, 215, 4, 181, 149, 129, 50, 54, 72,
    205, 154, 94, 232, 195, 182, 85, 114, 157, 219, 174, 13, 103, 14, 90, 147,
    95, 218, 75, 126, 215, 246, 26, 77, 192, 156, 12, 113, 20, 115, 152, 30,
    106, 53, 96, 97, 93, 20, 63, 146, 57, 190, 131, 253, 115, 123, 92, 210,
    14, 110, 220, 149, 230, 202, 63, 56, 85, 31, 51, 16, 4, 156, 31, 221,
    64, 138, 20, 91, 43, 46, 148, 169, 193, 202, 180, 21, 124, 84, 207, 114,
    17, 153, 104, 250, 46, 102, 127, 23, 54, 221, 163, 151, 198, 108, 243, 8,
    240, 175, 232, 128, 36, 116, 227, 152, 224, 170, 78, 107, 124, 185, 130, 59,
    222, 58, 65, 29, 241, 139, 118, 68, 69, 85, 189, 220, 41, 35, 111, 72,
    220, 237, 63, 76, 165, 176, 51, 116, 182, 19, 86, 58, 133, 56, 78, 196,
    211, 89, 123, 185, 4, 21, 199, 67, 174, 31, 173, 42, 184, 86, 167, 25,
    43, 176, 244, 212, 143, 13, 173, 140, 129, 224, 71, 23, 155, 211, 112, 96,
    228, 211, 37, 173, 41, 203, 243, 176, 101, 51, 229, 172, 168, 98, 160, 81,
    215, 251, 147, 254, 45, 45, 150, 210, 167, 144, 126, 145, 161, 102, 130, 120,
    61, 181, 80, 60, 240, 89, 169, 97, 138, 16, 48, 179, 209, 24, 178, 69,
    144, 236, 93, 105, 222, 34, 171, 252, 55, 229, 197, 150, 1, 89, 205, 91,
    170, 188, 180, 236, 223, 69, 56, 230, 118, 230, 108, 110, 19, 208, 61, 255,
    128, 87, 106, 40, 74, 249, 124, 247, 4, 129, 168, 41, 34, 19, 196, 59,
    126, 104, 184, 15, 181, 157, 75, 4, 52, 202, 49, 99, 121, 208, 132, 198,
    44, 248, 218, 158, 17, 239, 246, 72, 10, 130, 44, 105, 6, 199, 88, 149,
    88, 32, 5, 102, 16, 216, 9, 100, 114, 63, 222, 170, 34, 88, 159, 234,
    27, 186, 53, 109, 42, 210, 183, 28, 249, 109, 158, 5, 245, 223, 0, 164,
    26, 187, 127, 62, 199, 118, 169, 243, 75, 56, 169, 9, 71, 39, 146, 140,
    242, 146, 152, 40, 153, 238, 209, 179, 11, 206, 133, 242, 251, 67, 194, 161,
    40, 236, 158, 15, 109, 101, 39, 62, 125, 14, 62, 14, 59, 184, 250, 125,
    232, 204, 97, 125, 91, 149, 128, 250, 76, 29, 212, 185, 254, 173, 105, 59,
    198, 99, 4, 15, 149, 105, 71, 158, 255, 110, 229, 236, 59, 41, 247, 40,
    140, 9, 243, 62, 161, 242, 179, 39, 234, 170, 100, 101, 164, 5, 28, 109,
    25, 42, 67, 185, 96, 23, 140, 212, 72, 220, 68, 29, 196, 58, 116, 76,
    69, 179, 97, 60, 91, 150, 252, 105, 81, 172, 176, 173, 119, 145, 210, 254,
    30, 112, 77, 126, 209, 253, 146, 97, 72, 149, 216, 204, 147, 13, 114, 232,
    8, 151, 23, 250, 58, 107, 152, 128, 221, 15, 56, 149, 114, 21, 169, 91,
    248, 182, 212, 151, 194, 139, 186, 163, 137, 124, 80, 221, 189, 217, 12, 54,
    165, 239, 21, 8, 97, 163, 21, 243, 203, 122, 122, 108, 101, 106, 162, 198,
    113, 159, 104, 250, 71, 100, 27, 104, 123, 185, 217, 153, 138, 171, 197, 17,
    199, 44, 120, 23, 160, 235, 122, 127, 223, 145, 84, 102, 82, 71, 225, 46,
    25, 192, 2, 107, 135, 51, 109, 78, 206, 130, 15, 170, 170, 190, 229, 224,
    232, 137, 115, 59, 137, 145, 192, 130, 44, 145, 159, 185, 239, 122, 182, 124,
    72, 115, 99, 128, 155, 32, 53, 232, 131, 59, 121, 175, 188, 117, 209, 240,
    92, 232, 23, 152, 38, 177, 163, 227, 237, 137, 63, 116, 227, 23, 117, 36,
    236, 229, 181, 16, 251, 144, 211, 51, 175, 224, 89, 31, 237, 19, 58, 85,
    175, 170, 153, 221, 138, 202, 110, 31, 53, 190, 218, 156, 154, 219, 2, 4,
    151, 15, 112, 149, 179, 172, 42, 220, 8, 170, 98, 110, 204, 29, 236, 14,
    146, 206, 186, 56, 186, 109, 32, 63, 248, 130, 104, 202, 87, 129, 188, 230,
    218, 110, 192, 21, 206, 230, 28, 180, 191, 78, 33, 202, 180, 233, 39, 193,
    136, 135, 30, 114, 119, 167, 219, 207, 150, 51, 148, 85, 252, 128, 183, 124,
    227, 2, 83, 92, 59, 120, 183, 20, 141, 107, 204, 138, 93, 35, 27, 64,
    76, 25, 228, 169, 119, 146, 43, 148, 249, 192, 76, 46, 73, 224, 66, 43,
    48, 117, 22, 113, 47, 16, 109, 12, 67, 36, 133, 156, 70, 244, 22, 192,
    83, 228, 192, 30, 112, 186, 181, 152, 99, 23, 2, 115, 104, 120, 212, 20,
    147, 64, 70, 126, 187, 153, 249, 75, 172, 116, 117, 218, 108, 39, 178, 95,
    6, 237, 244, 77, 254, 95, 48, 26, 157, 137, 166, 246, 11, 112, 106, 215,
    37, 163, 133, 253, 154, 60, 5, 131, 154, 227, 240, 190, 224, 10, 156, 57,
    228, 38, 156, 210, 177, 145, 29, 92, 66, 178, 135, 123, 31, 160, 159, 115,
    77, 23, 67, 97, 3, 243, 158, 96, 127, 152, 5, 53, 184, 218, 180, 106,
    80, 93, 37, 147, 123, 244, 65, 63, 103, 232, 220, 20, 75, 96, 49, 94,
    96, 209, 109, 149, 16, 199, 223, 181, 41, 240, 241, 4, 230, 46, 70, 215,
    1, 188, 12, 73, 7, 214, 231, 54, 49, 117, 36, 50, 60, 200, 193, 129,
    250, 121, 47, 204, 216, 144, 245, 50, 3, 175, 168, 3, 141, 226, 213, 241,
    83, 131, 16, 216, 154, 3, 47, 187, 206, 59, 251, 179, 135, 155, 194, 70,
    134, 53, 22, 13, 83, 0, 164, 174, 128, 232, 205, 219, 56, 73, 210, 157,
    18, 155, 113, 147, 95, 126, 206, 90, 24, 32, 79, 14, 226, 235, 255, 103,
    136, 99, 48, 232, 131, 235, 0, 94, 41, 128, 67, 154, 175, 191, 137, 205,
    125, 72, 118, 114, 69, 101, 215, 85, 231, 115, 237, 183, 187, 211, 196, 195,
    111, 162, 106, 179, 213, 49, 32, 48, 0, 198, 60, 16, 230, 144, 51, 139,
    22, 101, 216, 69, 203, 139, 202, 178, 208, 107, 140, 24, 65, 55, 80, 209,
    18, 122, 11, 60, 241, 133, 89, 80, 62, 108, 189, 181, 65, 223, 200, 61,
    149, 203, 112, 97, 74, 195, 201, 169, 62, 122, 192, 89, 231, 244, 83, 240,
    33, 21, 166, 151, 141, 32, 102, 197, 94, 243, 52, 229, 118, 7, 177, 55,
    98, 97, 246, 91, 68, 102, 33, 205, 225, 163, 253, 89, 234, 120, 84, 1,
    246, 21, 157, 105, 226, 141, 6, 93, 132, 8, 14, 236, 100, 172, 153, 144,
    190, 165, 137, 252, 143, 64, 142, 171, 15, 239, 52, 35, 247, 213, 50, 222,
    162, 12, 48, 172, 168, 49, 216, 229, 132, 54, 11, 51, 110, 78, 28, 101,
    13, 189, 26, 81, 252, 24, 79, 160, 121, 217, 204, 98, 5, 117, 67, 168,
    82, 80, 255, 173, 3, 8, 17, 37, 37, 221, 82, 211, 27, 238, 10, 228,
    119, 137, 95, 177, 38, 68, 174, 243, 196, 125, 113, 203, 45, 43, 91, 42,
    94, 186, 244, 145, 3, 214, 100, 126, 235, 163, 177, 144, 108, 201, 145, 167,
    18, 194, 77, 120, 38, 161, 191, 130, 195, 139, 64, 102, 68, 24, 148, 161,
    81, 212, 6, 254, 175, 1, 9, 147, 171, 151, 26, 251, 107, 253, 42, 215,
    239, 248, 238, 210, 242, 148, 107, 150, 20, 182, 116, 45, 12, 176, 34, 45,
    233, 225, 178, 42, 245, 179, 63, 173, 48, 47, 122, 31, 208, 185, 200, 174,
    134, 82, 95, 26, 150, 194, 193, 167, 198, 71, 166, 86, 99, 77, 188, 184,
    156, 84, 108, 185, 50, 14, 114, 123, 98, 102, 24, 89, 35, 56, 86, 211,
    88, 223, 39, 67, 207, 19, 208, 199, 58, 145, 13, 21, 184, 173, 40, 4,
    189, 122, 117, 23, 217, 106, 171, 71, 233, 127, 35, 224, 19, 219, 29, 129,
    10, 160, 218, 96, 244, 120, 39, 112, 73, 235, 191, 211, 46, 199, 226, 155,
    7, 120, 172, 213, 144, 55, 222, 244, 201, 23, 253, 176, 78, 116, 238, 43,
    219, 199, 116, 140, 185, 77, 55, 173, 57, 44, 165, 13, 164, 140, 36, 143,
    87, 197, 249, 220, 47, 162, 88, 189, 136, 17, 18, 85, 255, 164, 214, 69,
    53, 138, 75, 72, 247, 161, 233, 111, 132, 171, 201, 35, 131, 141, 127, 41,
    221, 217, 187, 29, 205, 236, 34, 65, 17, 153, 160, 58, 93, 36, 142, 222,
    254, 123, 154, 68, 160, 38, 186, 118, 207, 185, 159, 139, 74, 143, 38, 242,
    214, 27, 70, 58, 250, 111, 234, 86, 99, 104, 33, 19, 61, 57, 92, 182,
    85, 71, 126, 196, 174, 131, 180, 78, 212, 100, 176, 56, 168, 44, 158, 133,
    217, 104, 148, 76, 121, 84, 159, 63, 36, 251, 79, 237, 2, 165, 135, 220,
    204, 113, 64, 116, 139, 191, 253, 52, 201, 159, 176, 176, 166, 13, 191, 165,
    0, 158, 183, 53, 159, 222, 9, 5, 194, 133, 209, 152, 146, 168, 169, 127,
    196, 168, 124, 106, 61, 108, 56, 180, 87, 44, 245, 218, 132, 49, 75, 184,
    210, 176, 252, 97, 115, 68, 140, 67, 169, 56, 151, 23, 114, 216, 146, 77,
    105, 212, 128, 96, 250, 62, 39, 158, 28, 100, 179, 243, 247, 229, 71, 4,
    207, 180, 148, 20, 31, 50, 62, 220, 130, 78, 227, 232, 242, 162, 127, 104,
    90, 174, 85, 94, 54, 181, 187, 50, 92, 131, 63, 96, 152, 12, 26, 75,
    130, 126, 150, 147, 160, 229, 243, 37, 205, 197, 171, 93, 178, 48, 169, 80,
    78, 86, 51, 63, 111, 189, 118, 65, 112, 71, 173, 244, 167, 173, 199, 123,
    79, 92, 219, 190, 135, 213, 242, 240, 61, 182, 81, 227, 40, 110, 134, 155,
    195, 61, 207, 57, 248, 223, 206, 156, 129, 57, 234, 10, 66, 78, 11, 224,
    245, 33, 119, 254, 124, 103, 161, 133, 125, 19, 251, 163, 30, 28, 194, 37,
    46, 99, 167, 131, 234, 65, 67, 5, 31, 104, 225, 60, 239, 101, 252, 154,
    239, 234, 30, 123, 66, 248, 40, 135, 200, 86, 197, 178, 152, 132, 152, 66,
    247, 250, 114, 115, 132, 61, 146, 159, 167, 111, 178, 160, 152, 54, 242, 31,
    213, 70, 136, 210, 178, 22, 180, 156, 236, 58, 27, 39, 197, 12, 209, 228,
    142, 214, 135, 92, 86, 200, 238, 29, 9, 28, 42, 145, 200, 51, 153, 177,
    226, 191, 39, 218, 29, 35, 171, 117, 36, 165, 55, 116, 238, 252, 222, 172,
    188, 166, 193, 95, 200, 122, 63, 178, 40, 13, 208, 39, 86, 24, 114, 108,
    158, 176, 180, 126, 92, 33, 234, 70, 190, 176, 253, 64, 135, 79, 159, 147,
    127, 201, 233, 75, 214, 18, 88, 249, 38, 159, 186, 153, 142, 160, 34, 187,
    84, 253, 33, 246, 177, 52, 102, 166, 50, 52, 142, 137, 153, 14, 5, 157,
    80, 139, 202, 216, 51, 60, 32, 106, 83, 191, 201, 112, 200, 189, 88, 11,
    148, 64, 191, 77, 145, 177, 100, 244, 34, 116, 107, 237, 42, 26, 9, 6,
    66, 204, 79, 26, 28, 11, 216, 48, 91, 113, 174, 95, 9, 81, 16, 254,
    158, 135, 28, 156, 156, 240, 5, 133, 49, 220, 65, 36, 180, 5, 158, 67,
    212, 2, 181, 192, 22, 133, 109, 22, 106, 165, 48, 244, 178, 117, 249, 70,
    255, 79, 206, 113, 210, 205, 164, 22, 196, 237, 215, 117, 137, 67, 0, 48,
    193, 36, 231, 12, 70, 241, 223, 109, 194, 251, 47, 16, 213, 168, 245, 47,
    182, 159, 192, 115, 229, 15, 109, 2, 117, 63, 163, 23, 209, 121, 53, 99,
    162, 121, 21, 21, 12, 80, 186, 212, 225, 84, 122, 120, 197, 217, 27, 104,
    183, 148, 219, 30, 39, 33, 28, 192, 66, 76, 43, 228, 27, 204, 183, 83,
    2, 218, 110, 152, 58, 89, 210, 181, 188, 104, 31, 186, 236, 98, 42, 112,
    103, 127, 66, 154, 157, 172, 99, 67, 231, 41, 24, 128, 206, 3, 49, 41,
    237, 131, 44, 206, 223, 123, 202, 137, 249, 141, 150, 191, 248, 8, 52, 175,
    57, 201, 73, 232, 58, 98, 198, 48, 201, 35, 212, 99, 207, 17, 132, 136,
    15, 132, 225, 94, 62, 62, 208, 0, 20, 171, 195, 235, 195, 37, 121, 132,
    90, 233, 76, 183, 224, 142, 15, 237, 54, 111, 6, 115, 11, 82, 184, 231,
    142, 247, 46, 195, 78, 90, 105, 196, 2, 161, 193, 211, 66, 80, 127, 187,
    10, 75, 249, 114, 10, 52, 4, 118, 221, 72, 81, 72, 234, 255, 75, 125,
    224, 53, 193, 85, 57, 192, 157, 215, 55, 138, 167, 101, 129, 173, 181, 69,
    143, 215, 43, 220, 240, 64, 105, 205, 8, 69, 140, 232, 242, 27, 44, 128,
    183, 255, 148, 157, 213, 245, 130, 210, 74, 164, 134, 73, 135, 238, 10, 57,
    238, 18, 119, 13, 190, 98, 149, 22, 203, 85, 254, 53, 219, 10, 88, 134,
    76, 136, 146, 174, 92, 231, 199, 165, 125, 18, 82, 1, 79, 228, 6, 83,
    25, 88, 228, 14, 155, 143, 158, 79, 214, 189, 94, 32, 134, 37, 218, 25,
    233, 47, 170, 103, 110, 241, 248, 255, 86, 162, 96, 235, 40, 221, 45, 226,
    136, 250, 76, 147, 227, 151, 72, 113, 217, 220, 7, 155, 81, 82, 246, 18,
    90, 87, 25, 90, 182, 84, 239, 207, 30, 197, 26, 126, 61, 211, 17, 95,
    45, 38, 27, 3, 5, 252, 110, 168, 12, 242, 107, 175, 68, 38, 108, 3,
    231, 74, 65, 241, 65, 203, 230, 213, 247, 143, 100, 226, 234, 68, 118, 141,
    73, 160, 93, 204, 131, 246, 255, 47, 166, 139, 1, 121, 168, 174, 61, 250,
    184, 207, 35, 50, 99, 150, 208, 245, 164, 185, 77, 144, 46, 244, 19, 216,
    87, 186, 204, 179, 154, 138, 237, 251, 73, 119, 93, 59, 252, 119, 180, 206,
    82, 160, 138, 70, 101, 97, 189, 194, 227, 38, 188, 155, 239, 137, 170, 135,
    134, 154, 141, 216, 60, 7, 230, 16, 187, 69, 175, 131, 226, 95, 56, 83,
    11, 254, 108, 187, 104, 112, 70, 47, 75, 123, 184, 3, 31, 177, 224, 154,
    155, 142, 183, 4, 133, 165, 41, 241, 1, 222, 209, 240, 161, 45, 168, 41,
    115, 93, 20, 215, 103, 195, 129, 70, 124, 36, 198, 46, 124, 43, 4, 230,
    164, 161, 59, 149, 254, 91, 59, 109, 113, 58, 147, 209, 119, 91, 196, 96,
    6, 17, 121, 181, 34, 74, 19, 223, 251, 153, 71, 199, 245, 42, 34, 16,
    90, 34, 83, 50, 37, 190, 40, 36, 188, 66, 225, 117, 194, 196, 106, 49,
    112, 205, 92, 129, 28, 169, 87, 193, 9, 236, 166, 200, 30, 125, 128, 60,
    74, 187, 227, 73, 164, 166, 247, 12, 29, 29, 9, 188, 251, 113, 124, 1,
    54, 221, 2, 54, 20, 201, 249, 231, 30, 217, 86, 214, 125, 203, 74, 7,
    241, 226, 3, 175, 50, 144, 204, 121, 70, 155, 179, 59, 187, 3, 216, 131,
    215, 46, 4, 240, 181, 198, 149, 209, 129, 200, 50, 117, 54, 214, 73, 188,
    205, 191, 154, 128, 94, 235, 232, 99, 195, 211, 182, 115, 85, 101, 114, 72,
    157, 73, 219, 232, 174, 0, 13, 53, 103, 235, 14, 32, 90, 126, 147, 155,
    95, 39, 218, 116, 75, 153, 126, 64, 215, 112, 246, 137, 26, 95, 77, 237,
    192, 244, 156, 36, 12, 16, 113, 117, 20, 120, 115, 23, 152, 186, 30, 228,
    106, 218, 53, 152, 96, 243, 97, 23, 94, 96, 20, 232, 63, 244, 147, 93,
    57, 10, 190, 227, 131, 60, 253, 163, 115, 160, 123, 178, 92, 145, 210, 38,
    14, 29, 110, 170, 220, 172, 149, 15, 230, 129, 202, 130, 63, 109, 56, 206,
    85, 19, 31, 224, 51, 144, 16, 229, 4, 219, 156, 190, 31, 202, 221, 170,
    64, 35, 138, 107, 20, 120, 91, 2, 43, 224, 46, 192, 148, 146, 169, 25,
    193, 233, 202, 78, 180, 230, 21, 110, 124, 127, 85, 51, 207, 167, 114, 135,
    17, 171, 153, 185, 104, 100, 250, 159, 46, 71, 102, 145, 127, 235, 23, 44,
    54, 217, 221, 124, 163, 139, 151, 182, 198, 106, 108, 122, 243, 163, 8, 239,
    240, 117, 175, 59, 232, 32, 128, 115, 36, 23, 116, 137, 227, 177, 152, 232,
    224, 190, 170, 130, 78, 51, 107, 192, 124, 122, 185, 145, 130, 145, 59, 137,
    222, 36, 58, 153, 65, 236, 29, 217, 242, 143, 139, 185, 118, 38, 68, 123,
    69, 164, 85, 17, 189, 162, 220, 197, 41, 140, 35, 171, 111, 161, 72, 138,
    220, 165, 237, 250, 63, 84, 76, 104, 165, 13, 176, 159, 52, 191, 116, 113,
    182, 57, 19, 104, 86, 111, 58, 27, 133, 44, 56, 100, 78, 130, 196, 71,
    211, 56, 89, 102, 123, 14, 185, 84, 4, 173, 21, 145, 199, 19, 67, 223,
    174, 185, 31, 46, 173, 179, 42, 225, 184, 77, 86, 71, 167, 194, 25, 82,
    43, 116, 176, 23, 244, 55, 213, 120, 143, 140, 13, 44, 173, 77, 140, 199,
    129, 219, 224, 127, 71, 106, 23, 122, 155, 199, 211, 235, 112, 120, 96, 160,
    228, 238, 211, 221, 37, 8, 173, 80, 42, 43, 203, 125, 243, 68, 176, 137,
    101, 78, 51, 54, 229, 49, 172, 12, 168, 116, 98, 217, 160, 24, 81, 189,
    215, 253, 251, 151, 147, 1, 254, 212, 45, 176, 45, 182, 150, 148, 210, 248,
    167, 201, 144, 163, 126, 214, 145, 186, 161, 24, 102, 139, 130, 161, 120, 194,
    61, 222, 181, 108, 80, 133, 60, 122, 240, 244, 89, 122, 169, 195, 97, 203,
    138, 144, 16, 198, 48, 49, 179, 162, 209, 55, 24, 106, 178, 139, 69, 101,
    144, 172, 236, 8, 93, 141, 105, 21, 222, 213, 34, 239, 171, 64, 252, 165,
    55, 7, 229, 243, 197, 32, 151, 21, 1, 120, 89, 163, 205, 102, 91, 97,
    170, 188, 188, 138, 180, 93, 237, 204, 223, 184, 69, 108, 56, 175, 230, 141,
    118, 100, 230, 64, 108, 107, 110, 27, 19, 77, 208, 35, 61, 1, 255, 93,
    128, 166, 87, 92, 106, 225, 40, 83, 74, 86, 249, 2, 124, 9, 247, 227,
    4, 198, 129, 20, 168, 208, 41, 183, 34, 71, 19, 121, 196, 147, 59, 59,
    126, 193, 105, 46, 184, 6, 15, 76, 181, 167, 157, 193, 75, 81, 4, 249,
    52, 150, 202, 44, 49, 24, 99, 66, 121, 194, 208, 225, 132, 212, 198, 73,
    44, 96, 248, 169, 218, 94, 158, 228, 17, 26, 239, 25, 246, 7, 72, 76,
    10, 134, 130, 148, 44, 140, 105, 43, 6, 82, 199, 146, 88, 254, 149, 119,
    88, 200, 32, 202, 5, 142, 102, 33, 16, 174, 216, 79, 9, 107, 100, 191,
    114, 208, 63, 193, 222, 55, 171, 39, 34, 185, 88, 233, 159, 253, 234, 180,
    27, 122, 186, 21, 53, 163, 109, 192, 42, 31, 210, 110, 183, 43, 28, 219,
    249, 48, 109, 181, 158, 65, 5, 28, 245, 47, 223, 230, 0, 215, 164, 206,
    26, 63, 187, 85, 127, 227, 62, 148, 199, 173, 118, 51, 169, 171, 243, 150,
    75, 245, 56, 124, 169, 209, 9, 183, 71, 179, 39, 128, 146, 151, 140, 252,
    242, 178, 146, 114, 152, 197, 40, 30, 153, 42, 238, 135, 209, 27, 180, 136,
    11, 234, 206, 207, 134, 81, 242, 219, 252, 225, 67, 167, 194, 251, 161, 119,
    40, 41, 236, 110, 158, 105, 15, 98, 109, 5, 101, 170, 39, 242, 62, 9,
    125, 184, 14, 14, 62, 101, 15, 236, 59, 173, 184, 29, 250, 149, 125, 204,
    232, 13, 204, 149, 97, 253, 126, 112, 91, 21, 149, 15, 128, 107, 250, 151,
    76, 58, 29, 220, 212, 23, 185, 42, 254, 145, 173, 172, 105, 150, 60, 179,
    198, 208, 99, 202, 4, 157, 15, 104, 149, 199, 105, 130, 71, 239, 158, 248,
    255, 208, 110, 230, 229, 69, 236, 188, 59, 19, 41, 129, 247, 249, 40, 87,
    140, 39, 9, 56, 243, 118, 62, 186, 161, 67, 242, 206, 179, 238, 40, 146,
    234, 88, 170, 63, 100, 216, 102, 32, 164, 223, 5, 109, 28, 210, 109, 186,
    25, 86, 42, 31, 67, 21, 185, 89, 96, 211, 23, 224, 140, 13, 212, 175,
    72, 35, 220, 85, 68, 139, 29, 58, 196, 56, 58, 19, 116, 176, 76, 237,
    69, 24, 179, 16, 97, 89, 60, 181, 91, 89, 150, 229, 252, 34, 105, 236,
    81, 98, 172, 51, 176, 203, 173, 211, 120, 102, 145, 144, 210, 45, 254, 251,
    30, 115, 113, 156, 77, 246, 126, 218, 209, 123, 253, 190, 146, 20, 97, 53,
    72, 50, 149, 4, 216, 179, 204, 2, 147, 14, 13, 219, 114, 182, 232, 154,
    8, 108, 151, 221, 23, 102, 250, 153, 59, 185, 107, 170, 152, 116, 128, 175,
    221, 156, 16, 31, 56, 202, 149, 110, 114, 84, 21, 202, 169, 46, 91, 138,
    248, 148, 182, 176, 212, 1, 151, 253, 194, 161, 139, 24, 186, 214, 163, 201,
    137, 68, 124, 43, 80, 8, 221, 238, 189, 24, 217, 116, 12, 49, 54, 78,
    165, 64, 239, 213, 21, 141, 8, 172, 97, 102, 163, 120, 21, 32, 243, 7,
    203, 195, 122, 244, 122, 133, 108, 222, 101, 139, 106, 55, 162, 49, 198, 144,
    113, 191, 159, 13, 104, 84, 250, 165, 71, 130, 100, 44, 27, 111, 104, 57,
    123, 38, 185, 143, 217, 236, 153, 36, 138, 161, 171, 140, 197, 162, 17, 164,
    199, 77, 44, 140, 120, 55, 23, 116, 160, 120, 235, 199, 122, 106, 127, 219,
    223, 18, 145, 173, 84, 14, 102, 56, 82, 194, 71, 77, 225, 179, 46, 185,
    25, 146, 192, 224, 2, 120, 107, 35, 135, 167, 51, 127, 109, 230, 78, 233,
    206, 109, 130, 129, 15, 172, 170, 29, 170, 202, 190, 219, 229, 144, 224, 19,
    232, 177, 137, 23, 115, 32, 59, 117, 137, 145, 145, 122, 192, 51, 130, 190,
    44, 235, 145, 71, 159, 100, 185, 171, 239, 163, 122, 106, 182, 139, 124, 217,
    72, 101, 115, 211, 99, 235, 128, 191, 155, 126, 32, 235, 53, 0, 232, 73,
    131, 3, 59, 155, 121, 144, 175, 226, 188, 214, 117, 200, 209, 198, 240, 46,
    92, 244, 232, 96, 23, 243, 152, 218, 38, 145, 178, 160, 163, 60, 227, 10,
    237, 95, 137, 112, 63, 153, 116, 39, 228, 186, 23, 120, 117, 16, 36, 244,
    236, 34, 229, 89, 181, 89, 16, 24, 251, 45, 144, 102, 211, 203, 51, 98,
    175, 13, 224, 211, 89, 21, 31, 86, 237, 176, 19, 56, 58, 139, 85, 35,
    175, 116, 170, 185, 153, 102, 221, 108, 138, 45, 202, 84, 110, 202, 31, 156,
    53, 20, 190, 123, 218, 246, 156, 115, 154, 182, 219, 14, 2, 179, 4, 50,
    151, 107, 15, 21, 112, 253, 149, 13, 179, 150, 172, 145, 42, 22, 220, 58,
    8, 242, 170, 5, 98, 105, 110, 41, 204, 149, 29, 173, 236, 101, 14, 184,
    146, 238, 206, 67, 186, 118, 56, 39, 186, 210, 109, 223, 32, 216, 63, 88,
    248, 239, 130, 199, 104, 157, 202, 208, 87, 249, 129, 19, 188, 69, 230, 207,
    219, 43, 110, 31, 192, 163, 21, 122, 206, 215, 230, 47, 28, 65, 180, 48,
    191, 107, 78, 174, 33, 142, 202, 200, 180, 253, 233, 185, 39, 55, 193, 208,
    136, 26, 135, 42, 30, 197, 114, 178, 119, 251, 167, 225, 219, 81, 207, 233,
    150, 171, 51, 173, 148, 226, 85, 63, 252, 151, 128, 179, 183, 209, 124, 245,
    227, 9, 2, 86, 83, 225, 92, 166, 59, 147, 121, 71, 183, 208, 20, 198,
    141, 175, 107, 184, 204, 93, 138, 188, 93, 1, 35, 77, 27, 107, 64, 100,
    76, 7, 25, 26, 228, 94, 169, 95, 119, 254, 146, 82, 43, 140, 148, 134,
    249, 81, 193, 167, 76, 6, 46, 193, 73, 212, 225, 194, 66, 24, 43, 150,
    48, 67, 117, 237, 22, 205, 113, 79, 47, 168, 16, 251, 109, 241, 12, 36,
    67, 5, 36, 220, 133, 240, 156, 135, 70, 117, 244, 165, 22, 133, 192, 2,
    83, 204, 228, 76, 192, 33, 30, 148, 112, 98, 186, 104, 181, 89, 152, 218,
    99, 121, 23, 63, 2, 15, 115, 159, 104, 217, 120, 84, 212, 80, 20, 121,
    147, 79, 64, 176, 70, 33, 126, 176, 187, 160, 153, 159, 249, 18, 75, 201,
    172, 252, 116, 165, 117, 35, 218, 191, 108, 24, 39, 13, 178, 122, 95, 166,
    6, 26, 237, 116, 244, 177, 77, 64, 254, 81, 95, 113, 48, 11, 26, 204,
    157, 13, 137, 52, 166, 52, 246, 253, 11, 189, 112, 191, 106, 60, 216, 139,
    37, 28, 163, 19, 133, 103, 253, 33, 154, 101, 60, 104, 5, 65, 131, 99,
    155, 110, 227, 182, 240, 213, 190, 92, 224, 78, 10, 57, 156, 223, 57, 61,
    228, 11, 39, 58, 156, 22, 210, 70, 177, 51, 145, 28, 29, 200, 92, 214,
    66, 132, 178, 86, 135, 247, 123, 234, 31, 54, 160, 111, 159, 61, 115, 250,
    77, 216, 23, 56, 67, 68, 97, 176, 3, 229, 243, 100, 158, 62, 96, 212,
    127, 168, 152, 133, 5, 222, 53, 158, 184, 49, 218, 44, 180, 108, 106, 168,
    80, 48, 93, 197, 37, 229, 147, 126, 123, 172, 244, 71, 65, 189, 63, 86,
    103, 162, 232, 78, 220, 50, 20, 180, 75, 12, 96, 131, 50, 181, 94, 174,
    96, 91, 209, 58, 109, 91, 149, 161, 16, 42, 199, 153, 223, 74, 181, 17,
    41, 45, 240, 222, 241, 165, 4, 142, 230, 43, 46, 36, 70, 195, 215, 93,
    1, 113, 188, 29, 12, 166, 73, 187, 7, 203, 214, 217, 231, 201, 54, 221,
    49, 196, 117, 65, 36, 190, 50, 34, 60, 125, 200, 236, 193, 169, 129, 205,
    250, 174, 121, 139, 47, 246, 204, 159, 216, 244, 144, 185, 245, 150, 50, 207,
    3, 38, 175, 242, 168, 252, 3, 38, 141, 68, 226, 143, 213, 203, 241, 74,
    83, 95, 131, 69, 16, 7, 216, 154, 154, 177, 3, 123, 47, 112, 187, 254,
    206, 119, 59, 119, 251, 138, 179, 186, 135, 137, 155, 38, 194, 97, 70, 160,
    134, 10, 53, 85, 22, 98, 13, 18, 83, 228, 1, 17, 164, 231, 174, 136,
    128, 27, 232, 69, 205, 64, 219, 215, 57, 238, 73, 164, 210, 245, 157, 255,
    18, 82, 155, 220, 113, 151, 147, 250, 95, 211, 126, 197, 207, 84, 90, 87,
    24, 37, 32, 189, 79, 143, 14, 88, 226, 221, 235, 162, 255, 241, 103, 47,
    136, 17, 99, 35, 48, 98, 232, 201, 131, 37, 235, 171, 0, 62, 94, 132,
    41, 3, 128, 41, 67, 172, 154, 127, 175, 8, 191, 141, 137, 123, 205, 131,
    125, 255, 72, 72, 118, 52, 114, 75, 69, 173, 101, 138, 215, 192, 85, 53,
    231, 82, 115, 111, 237, 142, 183, 233, 187, 80, 211, 161, 196, 90, 195, 247,
    111, 84, 162, 236, 106, 55, 179, 14, 213, 1, 49, 8, 32, 141, 48, 133,
    0, 234, 198, 144, 60, 243, 16, 153, 230, 120, 144, 172, 51, 32, 139, 100,
    22, 253, 101, 105, 216, 118, 69, 157, 203, 89, 139, 21, 202, 102, 178, 246,
    208, 225, 107, 93, 140, 94, 24, 6, 65, 163, 55, 141, 80, 197, 209, 226,
    18, 33, 122, 35, 11, 177, 60, 52, 241, 205, 133, 240, 89, 33, 80, 15,
    62, 68, 108, 222, 189, 229, 181, 50, 65, 103, 223, 213, 200, 22, 61, 247,
    150, 246, 203, 252, 112, 7, 97, 138, 74, 91, 195, 165, 201, 166, 169, 190,
    62, 98, 122, 171, 192, 52, 89, 142, 231, 97, 245, 64, 83, 151, 240, 143,
    33, 177, 22, 229, 166, 7, 151, 52, 141, 55, 32, 243, 102, 118, 197, 94,
    94, 118, 243, 55, 52, 7, 229, 177, 118, 7, 7, 7, 177, 7, 55, 118,
    98, 52, 97, 151, 246, 7, 91, 166, 68, 229, 103, 22, 33, 177, 205, 33,
    225, 94, 163, 197, 253, 118, 89, 102, 234, 243, 120, 32, 84, 55, 1, 141,
    246, 102, 21, 89, 157, 118, 105, 253, 226, 197, 141, 163, 6, 94, 93, 225,
    132, 141, 8, 1, 14, 55, 236, 84, 100, 32, 172, 120, 153, 243, 144, 234,
    190, 166, 165, 91, 138, 7, 252, 246, 143, 151, 64, 97, 142, 52, 171, 98,
    15, 33, 240, 205, 52, 177, 35, 33, 247, 22, 213, 103, 50, 229, 222, 68,
    162, 50, 12, 181, 48, 229, 172, 189, 168, 222, 49, 108, 216, 68, 229, 62,
    132, 247, 54, 61, 11, 22, 51, 200, 110, 213, 78, 223, 28, 103, 101, 65,
    13, 52, 189, 60, 26, 177, 81, 11, 252, 35, 24, 122, 79, 33, 160, 18,
    121, 15, 217, 80, 204, 33, 98, 89, 5, 240, 117, 133, 67, 205, 168, 241,
    82, 142, 80, 89, 255, 52, 173, 192, 3, 171, 8, 123, 17, 98, 37, 62,
    37, 143, 221, 240, 82, 151, 211, 83, 27, 64, 238, 245, 10, 97, 228, 231,
    119, 138, 137, 97, 95, 7, 177, 112, 38, 252, 68, 203, 174, 246, 243, 150,
    196, 190, 125, 169, 113, 166, 203, 201, 45, 165, 43, 195, 91, 91, 42, 74,
    95, 153, 186, 16, 244, 243, 145, 60, 3, 144, 214, 198, 100, 235, 126, 0,
    235, 100, 163, 139, 177, 32, 144, 51, 109, 172, 202, 144, 146, 120, 167, 230,
    18, 14, 194, 179, 77, 55, 120, 106, 38, 236, 161, 162, 191, 84, 130, 111,
    195, 133, 139, 49, 64, 141, 102, 32, 68, 8, 24, 49, 148, 1, 161, 214,
    81, 6, 212, 24, 6, 94, 254, 140, 175, 93, 1, 107, 9, 225, 147, 208,
    171, 226, 151, 209, 26, 197, 251, 80, 107, 141, 253, 55, 42, 163, 215, 65,
    239, 157, 248, 69, 238, 118, 210, 216, 242, 105, 148, 101, 107, 253, 150, 22,
    20, 246, 182, 178, 116, 102, 45, 202, 12, 21, 176, 139, 34, 89, 45, 203,
    233, 80, 225, 251, 178, 197, 42, 26, 245, 209, 179, 151, 63, 226, 173, 171,
    48, 65, 47, 215, 122, 163, 31, 43, 208, 55, 185, 253, 200, 142, 174, 107,
    134, 140, 82, 254, 95, 94, 26, 6, 150, 24, 194, 212, 193, 6, 167, 81,
    198, 208, 71, 147, 166, 225, 86, 9, 100, 107, 77, 1, 188, 93, 184, 175,
    156, 202, 84, 45, 108, 102, 185, 116, 50, 178, 14, 182, 114, 246, 123, 20,
    98, 203, 102, 45, 24, 89, 89, 34, 35, 139, 56, 176, 86, 21, 211, 12,
    88, 216, 223, 210, 39, 118, 67, 238, 207, 69, 19, 249, 208, 157, 199, 239,
    58, 22, 145, 150, 13, 253, 21, 107, 184, 101, 173, 149, 41, 105, 5, 242,
    190, 51, 122, 144, 117, 32, 23, 177, 217, 139, 106, 163, 171, 100, 71, 235,
    233, 230, 127, 167, 35, 120, 224, 146, 19, 144, 219, 202, 29, 172, 129, 109,
    10, 60, 160, 145, 218, 243, 96, 244, 244, 16, 120, 186, 39, 153, 112, 95,
    73, 0, 235, 126, 191, 235, 211, 100, 46, 198, 200, 214, 226, 144, 155, 3,
    7, 32, 120, 102, 172, 141, 213, 64, 144, 49, 55, 139, 222, 133, 244, 195,
    201, 214, 23, 161, 253, 1, 176, 148, 78, 49, 116, 24, 238, 8, 43, 68,
    219, 106, 199, 120, 116, 55, 140, 77, 185, 179, 77, 194, 55, 14, 173, 18,
    57, 111, 44, 130, 165, 84, 13, 191, 164, 162, 140, 161, 36, 236, 143, 38,
    87, 84, 197, 211, 250, 151, 220, 82, 47, 241, 162, 221, 88, 143, 189, 37,
    136, 231, 17, 228, 18, 97, 85, 10, 255, 245, 164, 238, 214, 64, 69, 27,
    53, 192, 138, 173, 75, 52, 72, 255, 247, 89, 161, 80, 233, 142, 111, 82,
    132, 62, 171, 37, 201, 98, 35, 17, 131, 123, 141, 8, 127, 172, 41, 3,
    221, 201, 217, 203, 187, 166, 29, 113, 205, 169, 236, 125, 34, 190, 65, 196,
    17, 74, 153, 42, 161, 91, 58, 91, 93, 195, 36, 43, 142, 165, 222, 45,
    254, 112, 123, 177, 154, 7, 68, 95, 160, 97, 38, 137, 186, 138, 119, 119,
    207, 150, 185, 243, 159, 246, 139, 174, 74, 203, 143, 68, 38, 252, 242, 38,
    214, 200, 28, 51, 70, 22, 58, 11, 250, 61, 111, 54, 234, 247, 86, 132,
    99, 65, 104, 101, 33, 103, 19, 28, 61, 223, 57, 78, 92, 213, 182, 110,
    85, 189, 71, 172, 126, 229, 197, 48, 174, 181, 131, 12, 180, 50, 78, 162,
    212, 62, 100, 229, 176, 68, 56, 216, 168, 108, 44, 49, 158, 222, 133, 168,
    218, 89, 104, 98, 148, 33, 76, 204, 121, 80, 84, 217, 159, 15, 63, 121,
    36, 241, 251, 168, 79, 205, 237, 67, 2, 133, 165, 117, 135, 240, 220, 5,
    204, 11, 113, 81, 64, 177, 116, 26, 139, 60, 191, 189, 253, 52, 52, 13,
    201, 18, 159, 160, 176, 33, 176, 79, 166, 122, 13, 24, 191, 35, 165, 252,
};


#define ERRLOC_SIZEX (18)

static short i2a(short a)
{
    short i, table_index;

    if ((a == 0) || (a >= 8192)) {
        return -1;
    }
    table_index = i_to_a[a] << 5;
    for (i = table_index; i < table_index + 32; i++) {
        if (a2i(i) == a) {
            return i;
        }
    }
    return -1;
}


static int bch_locate_errors(short *syn_val, short *perrlocations, int maxerrors)
{
    int *pMatrix_a;
    int *pMatrix_b;
    int *pMatrix_c;
    int *pElement;

    int _Matrix_a[19];
    int _Matrix_b[19];
    int _Matrix_c[20];
    int _Element[11];
    short _err_loc_ply[ERRLOC_SIZEX][44];

    const int bch_length = 104;
    const int data_length = 4096;
    const int block_length = 8191;

    int i, j, elp_sum ;
    int alpha, d_flg;
    int temp_index;

    short err_count;

    pMatrix_a = &_Matrix_a[0];
    pMatrix_b = &_Matrix_b[0];
    pMatrix_c = &_Matrix_c[0];
    pElement = &_Element[0];
    perr_loc_ply = &_err_loc_ply[0][0];

    d_flg = 0;
    err_count = 0;
    // initialise table entries
    for (i = 1; i <= NR_SYNDROMES; i++) {
        SET_SYNVAL(i, i2a(SYNVAL(i)));
    }

    SET_MATRIX_C(0, 0);
    SET_MATRIX_C(1, SYNVAL(1));
    *err_loc_ply(0,0) = 1;
    *err_loc_ply(1,0) = 1;
    for (i = 1; i < 16; i++) {
        *err_loc_ply(0,i) = 0;
        *err_loc_ply(1,i) = 0;
    }
    SET_MATRIX_A(0, 0);
    SET_MATRIX_A(1, 0);
    SET_MATRIX_B(0, -1);
    SET_MATRIX_B(1, 0);
    alpha = -1;

    do {
        // skip even loops
        alpha += 2;
        if (MATRIX_C(alpha) != -1) {
            temp_index = alpha - 2;
            if (temp_index < 0) {
                temp_index = 0;
            }
            while ((temp_index > 0) && (MATRIX_C(temp_index) == -1)) {
                temp_index = temp_index - 2;
            }
            if (temp_index < 0) {
                temp_index = 0;
            }

            if (temp_index > 0) {
                j = temp_index;
                do {
                    j = j - 2;
                    if (j < 0) {
                        j = 0;
                    }
                    if ((MATRIX_C(j) != -1) && (MATRIX_B(temp_index) < MATRIX_B(j))) {
                        temp_index = j;
                    }
                } while (j > 0);
            }

            if (MATRIX_A(alpha) > MATRIX_A(temp_index) + alpha - temp_index) {
                SET_MATRIX_A(alpha + 2, MATRIX_A(alpha));
            } else {
                SET_MATRIX_A(alpha + 2, MATRIX_A(temp_index) + alpha - temp_index);
            }

            for (i = 0; i < 16; i++) {
                *err_loc_ply(alpha + 2,i) = 0;
            }

            for (i = 0; i <= MATRIX_A(temp_index); i++) {
                if (*err_loc_ply(temp_index,i) != 0) {
                    *err_loc_ply(alpha + 2,i + alpha - temp_index) = a2i((MATRIX_C(alpha) + block_length - MATRIX_C(temp_index) + i2a(*err_loc_ply(temp_index,i))) % block_length);
                }
            }
            for (i = 0; i <= MATRIX_A(alpha); i++) {
                *err_loc_ply(alpha + 2,i) ^= *err_loc_ply(alpha,i);
            }
        }
        else {
            SET_MATRIX_A(alpha + 2, MATRIX_A(alpha));
            for (i = 0; i <= MATRIX_A(alpha); i++) {
                *err_loc_ply(alpha + 2,i) = *err_loc_ply(alpha,i);
            }
        }

        SET_MATRIX_B(alpha + 2, alpha + 1 - MATRIX_A(alpha + 2));

        // Form (alpha+2)th discrepancy.
        if (alpha < 16) {
            if (SYNVAL(alpha + 2) != -1) {
                SET_MATRIX_C(alpha + 2, a2i(SYNVAL(alpha + 2)));
            } else {
                SET_MATRIX_C(alpha + 2, 0);
            }

            for (i = 1; i <= MATRIX_A(alpha + 2); i++) {
                if ((SYNVAL(alpha + 2 - i) != -1) && (*err_loc_ply(alpha + 2,i) != 0)) {
                    SET_MATRIX_C(alpha + 2, MATRIX_C(alpha + 2) ^ a2i((SYNVAL(alpha + 2 - i) + i2a(*err_loc_ply(alpha + 2,i))) % block_length));
                }
            }
            SET_MATRIX_C(alpha + 2, i2a(MATRIX_C(alpha + 2)));
        }

    } while ((alpha < 15) && (MATRIX_A(alpha + 2) <= 8));

    alpha = alpha + 2;
    SET_MATRIX_A(15, MATRIX_A(alpha));
    if (MATRIX_A(15) <= 8) {

        for (i = 1; i <= MATRIX_A(15); i++) {
            SET_ELEMENT(i, i2a(*err_loc_ply(alpha,i)));
        }

        for (i = 1; i <= block_length; i++) {
            elp_sum = 1;
            for (j = 1; j <= MATRIX_A(15); j++) {
                if (ELEMENT(j) != -1) {
                    SET_ELEMENT(j, (ELEMENT(j) + j) % block_length);
                    elp_sum ^= a2i(ELEMENT(j));
                }
            }

            if (!elp_sum) {
                if (err_count >= maxerrors) {
                    /* More errors than caller can handle */
                    return -1;
                }
                perrlocations[err_count] = block_length - i;
                err_count++;
            }
        }

        if (err_count == MATRIX_A(15)) {
            d_flg = 1;
        }
    }

    if (d_flg == 0) {
        /* Page is uncorrectable */
        return -1;
    }

    for (i = 0; i < err_count; i++) {
        if (perrlocations[i] >= bch_length) {
            /* Return bit index of errors in 512 byte page */
            perrlocations[i] = bch_length + data_length - 1 -
                perrlocations[i];
        } else {
            /* Error in ECC area, these are flagged but ignored */
            perrlocations[i] = -1;
        }
    }
    return err_count;
}

static void fix_errors(int i, short bitindex, void *pmemstart)
{
	unsigned char *perrbyte;
	
	if (bitindex == -1) {
		/* Bit error was outside data area so we ignore it */
		return;
	}
	
	/* Access the erroneous byte */
	perrbyte = (unsigned char *)pmemstart + (i << 9) + (bitindex >> 3);

	/* The NAND controller is wired little endian so when the host does 
	 * a word copy from the FIFOs to memory we get an endian swap. Normally
	 * this makes no difference as the swaps on write and read cancel each
	 * other out, however we do need to account for it when working out 
	 * which byte to correct otherwise we correct byte 3 instead of byte 0,
	 * or byte 2 instead of byte 1.
	 */
	perrbyte = (unsigned char *)((unsigned long)perrbyte ^ 3);

	/* Flip the bit */
	*perrbyte ^= (1 << (bitindex & 7));
}


